# -*- coding: utf-8 -*-
#
# Copyright (c) 2004 Mikhail Kashkin (mailbox@xen.ru).
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.0 (ZPL).
""" Gear User Folder Module.

$Id: cvs string $
"""
import time
from thread import allocate_lock

import Globals
import zLOG
import OFS

from Globals import DTMLFile

from OFS.Folder import Folder
from OFS.PropertyManager import PropertyManager

from AccessControl import ClassSecurityInfo
from AccessControl.Role import RoleManager
from AccessControl.User import BasicUser, BasicUserFolder, SimpleUser

def sql_quote(v):
    if v.find("'") >= 0: return v.replace("'", "''")
    return v


# cache over Zope
cache_hash = {} # {oid -> { [cached_users, negative_cached_users] }

# we must lock before cleanup
cleanup_lock = allocate_lock()
# when we cleanup last time
global last_cleanup
last_cleanup = 0

debug = 0
   
class GearUserFolder(Folder, BasicUserFolder):
    """Gear UserFolder object."""

    meta_type='Gear User Folder'
    id       ='acl_users'
    title    ='Gear User Folder'

    isAnObjectManager=1
    isPrincipiaFolderish=1
    __allow_access_to_unprotected_subobjects__=1
    security = ClassSecurityInfo()
    
    # cleanup interval in seconds
    cleanup_interval = 600
    # cache check period in sec
    #check_interval = 300
    # to use full log, you can switch off cache. for more detales see doc.txt
    paranoid_log = 0 # full logging flag.
    # is passwords must bu encrypted
    encrypt_passwords = 0

    _properties = (
        {'id':'cleanup_interval', 'type':'int',},
        {'id':'encrypt_passwords', 'type':'boolean',},
        #{'id':'check_interval', 'type':'int',},
        #{'id':'default_roles', 'type':'tokens',},
        )
    
    manage_options=( \
        {'label':'Contents',   'action':'manage_main'},
        {'label':'Users',      'action':'manage_mainUsers'},
        {'label':'Properties', 'action':'manage_propertiesForm'},
        {'label':'Cache',      'action':'manage_cache'},
        {'label':'Security',   'action':'manage_access'},
        {'label':'Undo',       'action':'manage_UndoForm'},
        {'label':'Ownership',  'action':'manage_owner'},
      )
    
    manage_addUserForm = DTMLFile( 'www/addUser', globals() )
    manage_mainUsers=DTMLFile('www/mainUser', globals())
    manage_cache=DTMLFile('www/cache', globals())
    if debug: 
        manage_allcache=DTMLFile('www/allcache', globals())
    
    def __init__(self):
        pass
        
    def _getCache(self, userscache):
        """Returns cache for this object.
        """
        cache = cache_hash.get(self._p_oid, None)
        if cache is None:
            cache_hash[self._p_oid] = { \
                   'cached_users' : {},  # cached users : { login -> [timestamp, user object] }
                   'negative_cached_users' : {},  # negative cached users: { login }
                  }
        return cache_hash[self._p_oid][userscache]
    
    def _getCachedUser(self, name):
        """ check if user is in negative cache or try to 
        return User object from cache """
        now=time.time()
        self._cacheCleanup() # clear caches before
        negative_cached_users = self._getCache('negative_cached_users')
        if hasattr(negative_cached_users, name):
            negative_cached_users[name]=now
            return None
        cached_users = self._getCache('cached_users')
        try:
            userdata = cached_users[name]
            userdata[0]=now
            return userdata[1]
        except: 
            return None

    def _setCachedUser(self, name, object=None):
        """ set user object in cache, if `object` param 
        is None then add to negative cache """
        now = time.time()
        if object is not None:
            cached_users = self._getCache('cached_users')
            cached_users[name]=[now, object]
        else:
            negative_cached_users = self._getCache('negative_cached_users')
            negative_cached_users[name] = now

    def _clearCachedUser(self, name):
        """ delete user from all caches """
        cleanup_lock.acquire()
        cached_users = self._getCache('cached_users')
        negative_cached_users = self._getCache('negative_cached_users')
        try:
            del negative_cached_users[name]
        except:
            pass
        try:
            del cached_users[name]
        except:
            pass
        cleanup_lock.release()
        
    def _cacheCleanup(self):
        """ clear all old items """
        global last_cleanup
        now = time.time()
        cached_users = self._getCache('cached_users')
        negative_cached_users = self._getCache('negative_cached_users')
        if now - last_cleanup>self.cleanup_interval:
            timestamp = now + self.cleanup_interval
            # protect for threads
            cleanup_lock.acquire()
            try:
                for (key,value) in negative_cached_users.items():
                    if value < timestamp: del negative_cached_users[key]
                for (key,value) in cached_users.items():
                    if value[0] < timestamp: del cached_users[key]
            finally:
                cleanup_lock.release()

            last_cleanup = now
        else:        
            pass
            
    if debug:        
        def allUserCache(self, REQUEST=None, **ignore):
            return [x for x in self._getCache('cached_users')]
            #return cache_hash
            
    def getUserCache(self, REQUEST=None):
        """ list of cached users """
        cached_users = self._getCache('cached_users')
        return cached_users.keys()

    def getNegativeCache(self, REQUEST=None):
        """ list of users in negative cache """
        negative_cached_users = self._getCache('negative_cached_users')
        return negative_cached_users.keys()

    def getUserNames(self):
        """Return a list of usernames"""
        names=[]
        try:
            rez = self.recordValues()
        except:
            rez=None
        if rez:names = [x['name'] for x in rez]
        names.sort()
        return names

    def getUsers(self):
        """Return a list of user objects"""
        try:
            rez = self.recordValues()
            names = [x['name'] for x in rez]
        except:
            names = []
        return [self.getUser(n) for n in names]

    def getUser(self, name):
        """Return the named user object or None"""
        user = None
        if not self.paranoid_log:
            #check for paranoid log
            user = self._getCachedUser(name)
            if user:
                return user
        try:
            items=self.getRecord(name=name)
        except:
            return None
        try:
            # check item
            userdata = None
            if len(items)>0:
                item = items[0]
                userdata = {'name': item['name'], \
                            'password' : item['password'],\
                            'roles' : item['roles'].split(','),\
                            'domains' : ''}
        except:
            # XXX may be raise error? our source return non valid result.
            return None
        if userdata:
            user = GearUser(name=userdata['name'],\
                     password=userdata['password'],\
                     roles=userdata['roles'],\
                     domains=userdata['domains'])
        self._setCachedUser(name, user)
        return user

    def manage_cacheclear(self,REQUEST=None):
        """ Clear cache from ZMI """
        cleanup_lock.acquire()
        try:
            cached_users = {}
            negative_cached_users = {}
            last_cleanup = time.time()
        finally:        
            cleanup_lock.release()
        
        return self.manage_cache()

    def roleNormalize(self, roles):
        """ DB store roles as csv, this method convert to it """
        if type(roles) != type([]):
            if roles:
                roles=list(roles)
            else:
                roles=[]
                
        return ",".join(roles)

    def _doAddUser(self, name, password, roles, domains='', **kw):
        """Create a new user"""
        roles=self.roleNormalize(roles)
        if password is not None and self.encrypt_passwords:
            password = self._encryptPassword(password)
        self.addRecord(name=name,password=password,roles=roles,domains='', **kw)

    def _doChangeUser(self, name, password, roles, domains='', **kw):
        if roles is not None:
            roles=self.roleNormalize(roles)
            self.setRecord(name=name, roles=roles, domains='', **kw)
        if password is not None:
            if (  self.encrypt_passwords
                  and not self._isPasswordEncrypted(password)):
                password = self._encryptPassword(password)
            self.setPassword(password=password)

    def _doDelUsers(self, names):
        """ Source method must be undersand that `names` is sequence """
        self.delRecord(names=names)

Globals.default__class_init__(GearUserFolder)

class GearUser (SimpleUser):
    """
    A simple class for user objects.
    """
    def __init__(self, name,password,roles,domains):
        self.id=name
        self.name = name
        self.__=password
        self.roles=roles
        self.domains=domains
        
    def getUserName(self):
        """Return the username of a user"""
        return self.name

    def _getPassword(self):
        """Return the password of the user."""
        return self.__

    def getRoles(self):
        """Return the list of roles assigned to a user."""
        if self.name == 'Anonymous User': return tuple(self.roles)
        else: return tuple(self.roles) + ('Authenticated',)

    def getDomains(self):
        """Return the list of domain restrictions for a user"""
        return tuple(self.domains)
        
Globals.default__class_init__(GearUser)

def manage_addGearUserFolder(self, dtself=None, REQUEST=None, **ignored):
    """ adding a Gear User Folder"""
    f=GearUserFolder()
    self=self.this()
    try:    self._setObject('acl_users', f)
    except: return MessageDialog(
                   title  ='Item Exists',
                   message='This object already contains a User Folder',
                   action ='%s/manage_main' % REQUEST['URL1'])
    self.__allow_groups__=f
    #self.acl_users._post_init()

    if REQUEST is not None:
        REQUEST['RESPONSE'].redirect(self.absolute_url()+'/manage_main')
